using System;
using System.Windows.Forms;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace EQIA
{
    /// <summary>
    /// <para>An Xna Game object set up for accepting a System.Windows.Forms.Form as it's surface, it's functions are simple but
    /// lengthy. Providing the form to target is a simple matter of setting the TargetForm property, events will be hooked when
    /// a new form is set as the target, and events hooked to any previous form will be unhooked automatically.</para>
    /// <para>IsDrawOnForm: will allow the underlying game object to target the form during it's present calls, effectively replacing
    /// the form's background with the top-most back buffer.</para>
    /// <para>IsHidingGameWindow: will toggle wheather or not the Xna GameWindow is hidden or shown, when hidden it will automatically
    /// exit when the target form closes.</para>
    /// <para>IsCloseWithForm: will allow the Xna GameWindow to stay open if it's visible when the target form closes.</para>
    /// <para>IsFollowForm: will force the Xna GameWindow to follow the target form.</para>
    /// <para>IsActiveFixedTimeStep: will force the Updates to spread out across a fixed amount of time while the target form
    /// is the current focus of windows input.</para>
    /// <para>IsDeactiveFixedTimeStep: will force the Update to spread out across a fixed amount of time while the target form
    /// is not the current focus of windows input.</para>
    /// </summary>
    public class EQIA_Game : Game
    {
        #region Fields

        private GraphicsDeviceManager myGraphics;
        private Viewport myViewport;
        private Form myTarget;

        private string myTitle;

        private int myFrameRate, myActualFrameRate;
        private int myFrameTime;

        private bool isDrawForm;
        private bool isFormActive;
        private bool isActiveFixedTimeStep;
        private bool isDeactiveFixedTimeStep;
        private bool isCloseWithForm;
        private bool isHidingGameWindow;
        private bool isViewportInitialized;
        private bool isFollowForm;

        #endregion

        #region Properties

        /// <summary>
        /// The form that this game will be associated with. When setting this property, if a target form is
        /// already set, all events will be unhooked from that form and then hooked into the new target.
        /// </summary>
        public Form TargetForm
        {
            get { return myTarget; }
            set
            {
                if (myTarget != null)
                    UnhookFormEvents();

                myTarget = value;
                HookFormEvents();
            }
        }

        /// <summary>
        /// The frames per second of the current draw loop.
        /// </summary>
        public int FramesPerSecond { get { return myActualFrameRate; } }

        /// <summary>
        /// Should this game draw directly onto the form itself.
        /// </summary>
        public bool IsDrawOnForm
        {
            get { return isDrawForm; }
            set
            {
                if (!isDrawForm && value)
                {
                    isDrawForm = value;
                    try { myTarget.Invalidate(false); }
                    catch { }
                }
                isDrawForm = value;
            }
        }

        /// <summary>
        /// Should this game use it's TargetElapsedTime to space out Update calls when it's target form
        /// is the focus of windows input.
        /// </summary>
        public bool IsActiveFixedTimeStep
        {
            get { return isActiveFixedTimeStep; }
            set
            {
                isActiveFixedTimeStep = value;

                if (IsActive)
                    IsFixedTimeStep = isActiveFixedTimeStep;
            }
        }

        /// <summary>
        /// Should this game use it's TargetElapsedTime to space out Update class when it's target form
        /// is not the focus of windows input.
        /// </summary>
        public bool IsDeactiveFixedTimeStep
        {
            get { return isDeactiveFixedTimeStep; }
            set
            {
                isDeactiveFixedTimeStep = value;

                if (!IsActive)
                    IsFixedTimeStep = isDeactiveFixedTimeStep;
            }
        }

        /// <summary>
        /// Should the game close when it's target form closes.
        /// </summary>
        public bool IsCloseWithForm { get { return isCloseWithForm; } set { isCloseWithForm = value; } }

        /// <summary>
        /// Should the game window be hidden when a form is provided.
        /// </summary>
        public bool IsHidingGameWindow { get { return isHidingGameWindow; } set { isHidingGameWindow = value; } }

        /// <summary>
        /// Should the game window follow the target form, when this value is set to true, it will trigger the event
        /// handler for the form move to update the game window location.
        /// </summary>
        public bool IsFollowForm
        {
            get { return isFollowForm; }
            set
            {
                isFollowForm = value;

                if (isFollowForm && myTarget != null)
                    OnFormMove(null, null);
            }
        }

        /// <summary>
        /// Indicates wheather the game is currently the active application. Overriden to allow the target
        /// form to be factored into active checks.
        /// </summary>
        public new bool IsActive { get { return myTarget != null ? isFormActive : base.IsActive; } }

        #endregion

        #region Constructor

        /// <summary>
        /// Create a new instance of the TranceEngine game.
        /// </summary>
        public EQIA_Game()
            : base()
        {
            myGraphics = new GraphicsDeviceManager(this);
            myGraphics.SynchronizeWithVerticalRetrace = false;
            myGraphics.PreparingDeviceSettings += new System.EventHandler<PreparingDeviceSettingsEventArgs>(OnGraphicDeviceReset);
            Content.RootDirectory = "Content";

            myFrameRate = 0;
            myFrameTime = 0;

            isDrawForm = false;
            isFormActive = false;
            isActiveFixedTimeStep = false;
            isDeactiveFixedTimeStep = true;
            isCloseWithForm = true;
            isHidingGameWindow = true;
            isViewportInitialized = false;
            isFollowForm = false;

            IsFixedTimeStep = false;
        }

        #endregion

        #region Initialize, Update, and Draw Methods

        /// <summary>
        /// Initialize the game then hook into events that are required to handle the form.
        /// </summary>
        protected override void Initialize()
        {
            base.Initialize();
            ((Form)Control.FromHandle(Window.Handle)).Shown += new EventHandler(OnGameWindowShown);
        }

        /// <summary>
        /// Perform update logic on the game and then it's components.
        /// </summary>
        /// <param name="gameTime">Snapshot of the game time when Update was called.</param>
        protected override void Update(GameTime gameTime)
        {
            if (!isViewportInitialized)
            {
                try
                {
                    myViewport = GraphicsDevice.Viewport;
                    OnFormMove(null, null);
                    OnFormResize(null, null);
                    isViewportInitialized = true;
                }
                catch { isViewportInitialized = false; }
            }

            //Window.Title = myFrameRate.ToString();

            base.Update(gameTime);
        }

        /// <summary>
        /// Set the viewport, call Render, then draw the contents onto the target form 
        /// if one is provided and IsDrawOnForm is true.
        /// </summary>
        protected override sealed void Draw(GameTime gameTime)
        {
            // Framerate will be calculated by way of dual-identity average, this way the counter wount
            //  keep counting up then falling and we have a somewhat temporal dealiased count of frames.
            // It's also a good idea to never move this to the update loop since it is affected by the
            //  fixed time step variables.
            myFrameTime += gameTime.ElapsedRealTime.Milliseconds;
            if (myFrameTime >= 1000) // 1000 milliseconds = 1 second
            {
                myActualFrameRate = (myFrameRate + myActualFrameRate) / 2;
                myFrameRate = 0;
                myFrameTime = 0;
            }
            myFrameRate++;

            GraphicsDevice.Viewport = myViewport;
            GraphicsDevice.Clear(Color.CornflowerBlue);

            Render(gameTime);

            if (isDrawForm && myTarget != null && !myTarget.Disposing)
                GraphicsDevice.Present(new Rectangle(0, 0, myViewport.Width, myViewport.Height), null, myTarget.Handle);
        }

        /// <summary>
        /// Allow the game and it's components to be rendered onto the back buffer.
        /// </summary>
        protected virtual void Render(GameTime gameTime) { base.Draw(gameTime); }

        #endregion

        #region Event Hooking Methods

        /// <summary>
        /// Hook events onto the current form if the form is valid then trigger the Move and Resize
        /// event handlers to adjust the viewport and back buffer.
        /// </summary>
        private void HookFormEvents()
        {
            if (myTarget != null)
            {
                myTarget.Move += new EventHandler(OnFormMove);
                myTarget.Resize += new EventHandler(OnFormResize);
                myTarget.Activated += new EventHandler(OnFormActivated);
                myTarget.Deactivate += new EventHandler(OnFormDeactivated);
                myTarget.FormClosing += new FormClosingEventHandler(OnFormClosing);

                try
                {
                    myViewport = GraphicsDevice.Viewport;
                    OnFormMove(null, null);
                    OnFormResize(null, null);
                    isViewportInitialized = true;
                }
                catch (Exception) { isViewportInitialized = false; }
            }
        }

        /// <summary>
        /// Unhook events from the current form if the form is valid then reset the viewport with
        /// the GameWindow size.
        /// </summary>
        private void UnhookFormEvents()
        {
            if (myTarget != null)
            {
                myTarget.Move -= OnFormMove;
                myTarget.Resize -= OnFormResize;
                myTarget.Activated -= OnFormActivated;
                myTarget.Deactivate -= OnFormDeactivated;
                myTarget.FormClosing -= OnFormClosing;

                myViewport = GraphicsDevice.Viewport;
                myViewport.Width = Window.ClientBounds.Width;
                myViewport.Height = Window.ClientBounds.Height;
            }
        }

        #endregion

        #region Event Handlers

        /// <summary>
        /// In order to accomodate multiple present calls to multiple controls, the swap effect must be set to copy
        /// so that the back-buffer doesn't get discarded before the next control flips to screen space.
        /// </summary>
        private void OnGraphicDeviceReset(object sender, PreparingDeviceSettingsEventArgs e)
        {
            e.GraphicsDeviceInformation.PresentationParameters.SwapEffect = SwapEffect.Copy;
        }

        /// <summary>
        /// Need to catch when the game form is shown to hide it when there is a valid target form and IsHidingGameWindow
        /// is set to true.
        /// </summary>
        private void OnGameWindowShown(object sender, EventArgs e)
        {
            if (myTarget != null && isHidingGameWindow)
                ((Form)Control.FromHandle(Window.Handle)).Hide();
        }

        /// <summary>
        /// When the form moves, the xna window must move with it to ensure.
        /// </summary>
        private void OnFormMove(object sender, EventArgs e)
        {
            if (myTarget.Disposing || !isFollowForm) return;

            // Get the xna window as a form from it's handle
            Form xnaForm = (Form)Control.FromHandle(Window.Handle);

            // Get the location of the target form in screen co-ordinates, then offset it's
            //  location based on the location of the game window's client area in screen co-ordinates
            System.Drawing.Point location = myTarget.PointToScreen(myTarget.ClientRectangle.Location);
            System.Drawing.Point offset = xnaForm.PointToScreen(xnaForm.ClientRectangle.Location);
            offset.X -= xnaForm.Location.X;
            offset.Y -= xnaForm.Location.Y;

            // Set in the new location for the xna window
            xnaForm.Location = new System.Drawing.Point(location.X - offset.X, location.Y - offset.Y);
        }

        /// <summary>
        /// When the form resizes, the back buffer and viewport need to be resized to allow the form to fit
        /// inside, this will only increase the size of the back buffer since device resets can be costly on
        /// some machines.
        /// </summary>
        private void OnFormResize(object sender, EventArgs e)
        {
            if (myTarget.Disposing) return;

            try
            {
                bool reset = false;

                // Update the back buffer size so that the form will fit inside of it only if the form
                //  is larger than the current back buffer
                if (GraphicsDevice.PresentationParameters.BackBufferWidth < myTarget.ClientSize.Width)
                {
                    GraphicsDevice.PresentationParameters.BackBufferWidth = myTarget.ClientSize.Width;
                    myGraphics.PreferredBackBufferWidth = myTarget.ClientSize.Width;
                    reset = true;
                }
                if (GraphicsDevice.PresentationParameters.BackBufferHeight < myTarget.ClientSize.Height)
                {
                    GraphicsDevice.PresentationParameters.BackBufferHeight = myTarget.ClientSize.Height;
                    myGraphics.PreferredBackBufferHeight = myTarget.ClientSize.Height;
                    reset = true;
                }

                if (reset) myGraphics.GraphicsDevice.Reset(GraphicsDevice.PresentationParameters);

                // Update the viewport with the new back buffer size
                myViewport.Width = myTarget.ClientSize.Width;
                myViewport.Height = myTarget.ClientSize.Height;
            }
            catch { }
        }

        /// <summary>
        /// When the form becomes active set the internal flag and set the time step if IsActiveFixedTimeStep is true.
        /// </summary>
        private void OnFormActivated(object sender, EventArgs e)
        {
            isFormActive = true;

            if (IsActive)
                IsFixedTimeStep = IsActiveFixedTimeStep;
        }

        /// <summary>
        /// When the form becomes in-active set the internal flag and set the time step if IsDeactiveFixedTimeStep is true.
        /// </summary>
        private void OnFormDeactivated(object sender, EventArgs e)
        {
            isFormActive = false;

            if (IsActive)
                IsFixedTimeStep = IsDeactiveFixedTimeStep;
        }

        /// <summary>
        /// If IsCloseWithForm is true, this will close the game as well to prevent having the game hold the application
        /// open. If IsCloseWithForm is false, this will force the game window to show itself if IsHidingGameWindow is true
        /// and unhook any events associated with that form.
        /// </summary>
        private void OnFormClosing(object sender, FormClosingEventArgs e)
        {
            if (IsCloseWithForm)
                Exit();
            else
            {
                UnhookFormEvents();

                Form xnaForm = (Form)Control.FromHandle(Window.Handle);
                if (!xnaForm.Visible)
                    xnaForm.Show();
            }
        }

        #endregion
    }
}
